@@ -2,8 +2,8 @@
 
   array.c -
 
-  $Author: matz $
-  $Date: 2004/07/24 09:46:45 $
+  $Author: ocean $
+  $Date: 2005/11/15 07:38:38 $
   created at: Fri Aug  6 09:46:12 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -31,7 +31,7 @@ rb_mem_clear(mem, size)
     }
 }
 
-static void
+static inline void
 memfill(mem, size, val)
     register VALUE *mem;
     register long size;
@@ -50,7 +50,7 @@ rb_ary_modify_check(ary)
 {
     if (OBJ_FROZEN(ary)) rb_error_frozen("array");
     if (FL_TEST(ary, ARY_TMPLOCK))
-	rb_raise(rb_eTypeError, "can't modify array during iteration");
+	rb_raise(rb_eRuntimeError, "can't modify array during iteration");
     if (!OBJ_TAINTED(ary) && rb_safe_level() >= 4)
 	rb_raise(rb_eSecurityError, "Insecure: can't modify array");
 }
@@ -70,6 +70,7 @@ rb_ary_modify(ary)
 	RARRAY(ary)->ptr = ptr;
     }
 }
+
 VALUE
 rb_ary_freeze(ary)
     VALUE ary;
@@ -269,7 +270,6 @@ rb_ary_initialize(argc, argv, ary)
     long len;
     VALUE size, val;
 
-    rb_ary_modify(ary);
     if (rb_scan_args(argc, argv, "02", &size, &val) == 0) {
 	RARRAY(ary)->len = 0;
 	if (rb_block_given_p()) {
@@ -293,6 +293,7 @@ rb_ary_initialize(argc, argv, ary)
     if (len > 0 && len * (long)sizeof(VALUE) <= len) {
 	rb_raise(rb_eArgError, "array size too big");
     }
+    rb_ary_modify(ary);
     if (len > RARRAY(ary)->aux.capa) {
 	REALLOC_N(RARRAY(ary)->ptr, VALUE, len);
 	RARRAY(ary)->aux.capa = len;
@@ -304,7 +305,7 @@ rb_ary_initialize(argc, argv, ary)
 	    rb_warn("block supersedes default value argument");
 	}
 	for (i=0; i<len; i++) {
-	    RARRAY(ary)->ptr[i] = rb_yield(LONG2NUM(i));
+	    rb_ary_store(ary, i, rb_yield(LONG2NUM(i)));
 	    RARRAY(ary)->len = i + 1;
 	}
     }
@@ -333,9 +334,6 @@ rb_ary_s_create(argc, argv, klass)
 {
     VALUE ary = ary_alloc(klass);
 
-    if (argc < 0) {
-	rb_raise(rb_eArgError, "negative number of arguments");
-    }
     if (argc > 0) {
 	RARRAY(ary)->ptr = ALLOC_N(VALUE, argc);
 	MEMCPY(RARRAY(ary)->ptr, argv, VALUE, argc);
@@ -351,7 +349,6 @@ rb_ary_store(ary, idx, val)
     long idx;
     VALUE val;
 {
-    rb_ary_modify(ary);
     if (idx < 0) {
 	idx += RARRAY(ary)->len;
 	if (idx < 0) {
@@ -360,6 +357,7 @@ rb_ary_store(ary, idx, val)
 	}
     }
 
+    rb_ary_modify(ary);
     if (idx >= RARRAY(ary)->aux.capa) {
 	long new_capa = RARRAY(ary)->aux.capa / 2;
 
@@ -458,7 +456,7 @@ rb_ary_pop(ary)
     return RARRAY(ary)->ptr[--RARRAY(ary)->len];
 }
 
-static void
+static VALUE
 ary_make_shared(ary)
     VALUE ary;
 {
@@ -471,6 +469,11 @@ ary_make_shared(ary)
 	shared->aux.capa = RARRAY(ary)->aux.capa;
 	RARRAY(ary)->aux.shared = (VALUE)shared;
 	FL_SET(ary, ELTS_SHARED);
+	OBJ_FREEZE(shared);
+	return (VALUE)shared;
+    }
+    else {
+	return RARRAY(ary)->aux.shared;
     }
 }
 
@@ -546,9 +549,6 @@ rb_ary_unshift_m(argc, argv, ary)
 {
     long len = RARRAY(ary)->len;
 
-    if (argc < 0) {
-	rb_raise(rb_eArgError, "negative number of arguments");
-    }
     if (argc == 0) return ary;
 
     /* make rooms by setting the last item */
@@ -561,21 +561,28 @@ rb_ary_unshift_m(argc, argv, ary)
     return ary;
 }
 
+/* faster version - use this if you don't need to treat negative offset */
+static inline VALUE
+rb_ary_elt(ary, offset)
+    VALUE ary;
+    long offset;
+{
+    if (RARRAY(ary)->len == 0) return Qnil;
+    if (offset < 0 || RARRAY(ary)->len <= offset) {
+	return Qnil;
+    }
+    return RARRAY(ary)->ptr[offset];
+}
+
 VALUE
 rb_ary_entry(ary, offset)
     VALUE ary;
     long offset;
 {
-    if (RARRAY(ary)->len == 0) return Qnil;
-
     if (offset < 0) {
 	offset += RARRAY(ary)->len;
     }
-    if (offset < 0 || RARRAY(ary)->len <= offset) {
-	return Qnil;
-    }
-
-    return RARRAY(ary)->ptr[offset];
+    return rb_ary_elt(ary, offset);
 }
 
 static VALUE
@@ -583,7 +590,8 @@ rb_ary_subseq(ary, beg, len)
     VALUE ary;
     long beg, len;
 {
-    VALUE klass, ary2;
+    VALUE klass, ary2, shared;
+    VALUE *ptr;
 
     if (beg > RARRAY(ary)->len) return Qnil;
     if (beg < 0 || len < 0) return Qnil;
@@ -596,11 +604,12 @@ rb_ary_subseq(ary, beg, len)
     klass = rb_obj_class(ary);
     if (len == 0) return ary_new(klass, 0);
 
-    ary_make_shared(ary);
+    shared = ary_make_shared(ary);
+    ptr = RARRAY(ary)->ptr;
     ary2 = ary_alloc(klass);
-    RARRAY(ary2)->ptr = RARRAY(ary)->ptr + beg;
+    RARRAY(ary2)->ptr = ptr + beg;
     RARRAY(ary2)->len = len;
-    RARRAY(ary2)->aux.shared = RARRAY(ary)->aux.shared;
+    RARRAY(ary2)->aux.shared = shared;
     FL_SET(ary2, ELTS_SHARED);
 
     return ary2;
@@ -706,12 +715,16 @@ rb_ary_at(ary, pos)
 /*
  *  call-seq:
  *     array.first   ->   obj or nil
- *  
- *  Returns the first element of the array. If the array is empty,
- *  returns <code>nil</code>.
- *     
+ *     array.first(n) -> an_array
+ *
+ *  Returns the first element, or the first +n+ elements, of the array.
+ *  If the array is empty, the first form returns <code>nil</code>, and the
+ *  second form returns an empty array.
+ *
  *     a = [ "q", "r", "s", "t" ]
- *     a.first   #=> "q"
+ *     a.first    #=> "q"
+ *     a.first(1) #=> ["q"]
+ *     a.first(3) #=> ["q", "r", "s"]
  */
 
 static VALUE
@@ -873,6 +886,10 @@ rb_ary_rindex(ary, val)
     long i = RARRAY(ary)->len;
 
     while (i--) {
+	if (i > RARRAY(ary)->len) {
+	    i = RARRAY(ary)->len;
+	    continue;
+	}
 	if (rb_equal(RARRAY(ary)->ptr[i], val))
 	    return LONG2NUM(i);
     }
@@ -919,7 +936,7 @@ rb_ary_to_ary(obj)
 }
 
 static void
-rb_ary_update(ary, beg, len, rpl)
+rb_ary_splice(ary, beg, len, rpl)
     VALUE ary;
     long beg, len;
     VALUE rpl;
@@ -938,7 +955,6 @@ rb_ary_update(ary, beg, len, rpl)
 	len = RARRAY(ary)->len - beg;
     }
 
-    rb_ary_modify(ary);
     if (NIL_P(rpl)) {
 	rlen = 0;
     }
@@ -946,6 +962,7 @@ rb_ary_update(ary, beg, len, rpl)
 	rpl = rb_ary_to_ary(rpl);
 	rlen = RARRAY(rpl)->len;
     }
+    rb_ary_modify(ary);
 
     if (beg >= RARRAY(ary)->len) {
 	len = beg + rlen;
@@ -1026,7 +1043,7 @@ rb_ary_aset(argc, argv, ary)
 	if (SYMBOL_P(argv[1])) {
 	    rb_raise(rb_eTypeError, "Symbol as subarray length");
 	}
-	rb_ary_update(ary, NUM2LONG(argv[0]), NUM2LONG(argv[1]), argv[2]);
+	rb_ary_splice(ary, NUM2LONG(argv[0]), NUM2LONG(argv[1]), argv[2]);
 	return argv[2];
     }
     if (argc != 2) {
@@ -1041,7 +1058,7 @@ rb_ary_aset(argc, argv, ary)
     }
     if (rb_range_beg_len(argv[0], &beg, &len, RARRAY(ary)->len, 1)) {
 	/* check if idx is Range */
-	rb_ary_update(ary, beg, len, argv[1]);
+	rb_ary_splice(ary, beg, len, argv[1]);
 	return argv[1];
     }
 
@@ -1071,6 +1088,7 @@ rb_ary_insert(argc, argv, ary)
 {
     long pos;
 
+    if (argc == 1) return ary;
     if (argc < 1) {
 	rb_raise(rb_eArgError, "wrong number of arguments (at least 1)");
     }
@@ -1078,12 +1096,10 @@ rb_ary_insert(argc, argv, ary)
     if (pos == -1) {
 	pos = RARRAY(ary)->len;
     }
-    else if (pos < 0) {
+    if (pos < 0) {
 	pos++;
     }
-
-    if (argc == 1) return ary;
-    rb_ary_update(ary, pos, 0, rb_ary_new4(argc - 1, argv + 1));
+    rb_ary_splice(ary, pos, 0, rb_ary_new4(argc - 1, argv + 1));
     return ary;
 }
 
@@ -1529,40 +1545,69 @@ rb_ary_reverse_m(ary)
     return rb_ary_reverse(rb_ary_dup(ary));
 }
 
-static int
-sort_1(a, b)
-    VALUE *a, *b;
+struct ary_sort_data {
+    VALUE ary;
+    VALUE *ptr;
+    long len;
+};
+
+static void
+ary_sort_check(data)
+    struct ary_sort_data *data;
 {
-    VALUE retval = rb_yield_values(2, *a, *b);
-    return rb_cmpint(retval, *a, *b);
+    if (RARRAY(data->ary)->ptr != data->ptr || RARRAY(data->ary)->len != data->len) {
+	rb_raise(rb_eArgError, "array modified during sort");
+    }
 }
 
 static int
-sort_2(ap, bp)
+sort_1(a, b, data)
+    VALUE *a, *b;
+    struct ary_sort_data *data;
+{
+    VALUE retval = rb_yield_values(2, *a, *b);
+    int n;
+
+    n = rb_cmpint(retval, *a, *b);
+    ary_sort_check(data);
+    return n;
+}
+
+static int
+sort_2(ap, bp, data)
     VALUE *ap, *bp;
+    struct ary_sort_data *data;
 {
     VALUE retval;
-    long a = (long)*ap, b = (long)*bp;
+    VALUE a = *ap, b = *bp;
+    int n;
 
     if (FIXNUM_P(a) && FIXNUM_P(b)) {
-	if (a > b) return 1;
-	if (a < b) return -1;
+	if ((long)a > (long)b) return 1;
+	if ((long)a < (long)b) return -1;
 	return 0;
     }
-    if (TYPE(a) == T_STRING && TYPE(b) == T_STRING) {
-	return rb_str_cmp(a, b);
+    if (TYPE(a) == T_STRING) {
+	if (TYPE(b) == T_STRING) return rb_str_cmp(a, b);
     }
 
     retval = rb_funcall(a, id_cmp, 1, b);
-    return rb_cmpint(retval, a, b);
+    n = rb_cmpint(retval, a, b);
+    ary_sort_check(data);
+
+    return n;
 }
 
 static VALUE
 sort_internal(ary)
     VALUE ary;
 {
+    struct ary_sort_data data;
+
+    data.ary = ary;
+    data.ptr = RARRAY(ary)->ptr; data.len = RARRAY(ary)->len;
     qsort(RARRAY(ary)->ptr, RARRAY(ary)->len, sizeof(VALUE),
-	  rb_block_given_p()?sort_1:sort_2);
+	  rb_block_given_p()?sort_1:sort_2, &data);
     return ary;
 }
 
@@ -1681,7 +1726,7 @@ rb_ary_collect_bang(ary)
 
     rb_ary_modify(ary);
     for (i = 0; i < RARRAY(ary)->len; i++) {
-	RARRAY(ary)->ptr[i] = rb_yield(RARRAY(ary)->ptr[i]);
+	rb_ary_store(ary, i, rb_yield(RARRAY(ary)->ptr[i]));
     }
     return ary;
 }
@@ -1757,21 +1802,16 @@ rb_ary_values_at(argc, argv, ary)
  */
 
 static VALUE
-rb_ary_select(argc, argv, ary)
-    int argc;
-    VALUE *argv;
+rb_ary_select(ary)
     VALUE ary;
 {
     VALUE result;
     long i;
 
-    if (argc > 0) {
-	rb_raise(rb_eArgError, "wrong number of arguments (%d for 0)", argc);
-    }
     result = rb_ary_new2(RARRAY(ary)->len);
     for (i = 0; i < RARRAY(ary)->len; i++) {
 	if (RTEST(rb_yield(RARRAY(ary)->ptr[i]))) {
-	    rb_ary_push(result, RARRAY(ary)->ptr[i]);
+	    rb_ary_push(result, rb_ary_elt(ary, i));
 	}
     }
     return result;
@@ -1801,11 +1841,12 @@ rb_ary_delete(ary, item)
 {
     long i1, i2;
 
-    rb_ary_modify(ary);
     for (i1 = i2 = 0; i1 < RARRAY(ary)->len; i1++) {
-	if (rb_equal(RARRAY(ary)->ptr[i1], item)) continue;
+	VALUE e = RARRAY(ary)->ptr[i1];
+
+	if (rb_equal(e, item)) continue;
 	if (i1 != i2) {
-	    RARRAY(ary)->ptr[i2] = RARRAY(ary)->ptr[i1];
+	    rb_ary_store(ary, i2, e);
 	}
 	i2++;
     }
@@ -1816,11 +1857,14 @@ rb_ary_delete(ary, item)
 	return Qnil;
     }
 
-    RARRAY(ary)->len = i2;
-    if (i2 * 2 < RARRAY(ary)->aux.capa &&
+    rb_ary_modify(ary);
+    if (RARRAY(ary)->len > i2) {
+	RARRAY(ary)->len = i2;
+	if (i2 * 2 < RARRAY(ary)->aux.capa &&
 	    RARRAY(ary)->aux.capa > ARY_DEFAULT_SIZE) {
-	REALLOC_N(RARRAY(ary)->ptr, VALUE, i2 * 2);
-	RARRAY(ary)->aux.capa = i2 * 2;
+	    REALLOC_N(RARRAY(ary)->ptr, VALUE, i2 * 2);
+	    RARRAY(ary)->aux.capa = i2 * 2;
+	}
     }
 
     return item;
@@ -1834,13 +1878,13 @@ rb_ary_delete_at(ary, pos)
     long i, len = RARRAY(ary)->len;
     VALUE del;
 
-    rb_ary_modify(ary);
     if (pos >= len) return Qnil;
     if (pos < 0) {
 	pos += len;
 	if (pos < 0) return Qnil;
     }
 
+    rb_ary_modify(ary);
     del = RARRAY(ary)->ptr[pos];
     for (i = pos + 1; i < len; i++, pos++) {
 	RARRAY(ary)->ptr[pos] = RARRAY(ary)->ptr[i];
@@ -1905,7 +1949,6 @@ rb_ary_slice_bang(argc, argv, ary)
     VALUE arg1, arg2;
     long pos, len;
 
-    rb_ary_modify(ary);
     if (rb_scan_args(argc, argv, "11", &arg1, &arg2) == 2) {
 	pos = NUM2LONG(arg1);
 	len = NUM2LONG(arg2);
@@ -1914,7 +1957,7 @@ rb_ary_slice_bang(argc, argv, ary)
 	    pos = RARRAY(ary)->len + pos;
 	}
 	arg2 = rb_ary_subseq(ary, pos, len);
-	rb_ary_update(ary, pos, len, Qnil);	/* Qnil/rb_ary_new2(0) */
+	rb_ary_splice(ary, pos, len, Qnil);	/* Qnil/rb_ary_new2(0) */
 	return arg2;
     }
 
@@ -1943,14 +1986,16 @@ rb_ary_reject_bang(ary)
 
     rb_ary_modify(ary);
     for (i1 = i2 = 0; i1 < RARRAY(ary)->len; i1++) {
-	if (RTEST(rb_yield(RARRAY(ary)->ptr[i1]))) continue;
+	VALUE v = RARRAY(ary)->ptr[i1];
+	if (RTEST(rb_yield(v))) continue;
 	if (i1 != i2) {
-	    RARRAY(ary)->ptr[i2] = RARRAY(ary)->ptr[i1];
+	    rb_ary_store(ary, i2, v);
 	}
 	i2++;
     }
     if (RARRAY(ary)->len == i2) return Qnil;
-    RARRAY(ary)->len = i2;
+    if (i2 < RARRAY(ary)->len)
+	RARRAY(ary)->len = i2;
 
     return ary;
 }
@@ -2030,9 +2075,9 @@ rb_ary_zip(argc, argv, ary)
 	for (i=0; i<RARRAY(ary)->len; i++) {
 	    VALUE tmp = rb_ary_new2(argc+1);
 
-	    rb_ary_push(tmp, rb_ary_entry(ary, i));
+	    rb_ary_push(tmp, rb_ary_elt(ary, i));
 	    for (j=0; j<argc; j++) {
-		rb_ary_push(tmp, rb_ary_entry(argv[j], i));
+		rb_ary_push(tmp, rb_ary_elt(argv[j], i));
 	    }
 	    rb_yield(tmp);
 	}
@@ -2043,9 +2088,9 @@ rb_ary_zip(argc, argv, ary)
     for (i=0; i<len; i++) {
 	VALUE tmp = rb_ary_new2(argc+1);
 
-	rb_ary_push(tmp, rb_ary_entry(ary, i));
+	rb_ary_push(tmp, rb_ary_elt(ary, i));
 	for (j=0; j<argc; j++) {
-	    rb_ary_push(tmp, rb_ary_entry(argv[j], i));
+	    rb_ary_push(tmp, rb_ary_elt(argv[j], i));
 	}
 	rb_ary_push(result, tmp);
     }
@@ -2073,7 +2118,7 @@ rb_ary_transpose(ary)
     alen = RARRAY(ary)->len;
     if (alen == 0) return rb_ary_dup(ary);
     for (i=0; i<alen; i++) {
-	tmp = to_ary(RARRAY(ary)->ptr[i]);
+	tmp = to_ary(rb_ary_elt(ary, i));
 	if (elen < 0) {		/* first element */
 	    elen = RARRAY(tmp)->len;
 	    result = rb_ary_new2(elen);
@@ -2082,11 +2127,11 @@ rb_ary_transpose(ary)
 	    }
 	}
 	else if (elen != RARRAY(tmp)->len) {
-	    rb_raise(rb_eIndexError, "element size differ (%d should be %d)",
+	    rb_raise(rb_eIndexError, "element size differs (%d should be %d)",
 		     RARRAY(tmp)->len, elen);
 	}
 	for (j=0; j<elen; j++) {
-	    rb_ary_store(RARRAY(result)->ptr[j], i, RARRAY(tmp)->ptr[j]);
+	    rb_ary_store(rb_ary_elt(result, j), i, rb_ary_elt(tmp, j));
 	}
     }
     return result;
@@ -2108,15 +2153,17 @@ static VALUE
 rb_ary_replace(copy, orig)
     VALUE copy, orig;
 {
+    VALUE shared;
+
     rb_ary_modify(copy);
     orig = to_ary(orig);
     if (copy == orig) return copy;
-    ary_make_shared(orig);
+    shared = ary_make_shared(orig);
     if (RARRAY(copy)->ptr && !FL_TEST(copy, ELTS_SHARED))
 	free(RARRAY(copy)->ptr);
     RARRAY(copy)->ptr = RARRAY(orig)->ptr;
     RARRAY(copy)->len = RARRAY(orig)->len;
-    RARRAY(copy)->aux.shared = RARRAY(orig)->aux.shared;
+    RARRAY(copy)->aux.shared = shared;
     FL_SET(copy, ELTS_SHARED);
 
     return copy;
@@ -2214,9 +2261,7 @@ rb_ary_fill(argc, argv, ary)
 	    REALLOC_N(RARRAY(ary)->ptr, VALUE, end);
 	    RARRAY(ary)->aux.capa = end;
 	}
-	if (beg > RARRAY(ary)->len) {
-	    rb_mem_clear(RARRAY(ary)->ptr + RARRAY(ary)->len, end - RARRAY(ary)->len);
-	}
+	rb_mem_clear(RARRAY(ary)->ptr + RARRAY(ary)->len, end - RARRAY(ary)->len);
 	RARRAY(ary)->len = end;
     }
 
@@ -2224,7 +2269,7 @@ rb_ary_fill(argc, argv, ary)
 	VALUE v;
 	long i;
 
-	for (i=beg; i<RARRAY(ary)->len; i++) {
+	for (i=beg; i<end; i++) {
 	    v = rb_yield(LONG2NUM(i));
 	    if (i>=RARRAY(ary)->len) break;
 	    RARRAY(ary)->ptr[i] = v;
@@ -2282,7 +2327,7 @@ rb_ary_concat(x, y)
 {
     y = to_ary(y);
     if (RARRAY(y)->len > 0) {
-	rb_ary_update(x, RARRAY(x)->len, 0, y);
+	rb_ary_splice(x, RARRAY(x)->len, 0, y);
     }
     return x;
 }
@@ -2316,7 +2361,7 @@ rb_ary_times(ary, times)
     }
 
     len = NUM2LONG(times);
-    if (len == 0) return rb_ary_new2(0);
+    if (len == 0) return ary_new(rb_obj_class(ary), 0);
     if (len < 0) {
 	rb_raise(rb_eArgError, "negative argument");
     }
@@ -2360,17 +2405,15 @@ VALUE
 rb_ary_assoc(ary, key)
     VALUE ary, key;
 {
-    VALUE *p, *pend;
+    long i;
+    VALUE v;
 
-    p = RARRAY(ary)->ptr;
-    pend = p + RARRAY(ary)->len;
-    
-    while (p < pend) {
-	if (TYPE(*p) == T_ARRAY &&
-		RARRAY(*p)->len > 0 &&
-		rb_equal(RARRAY(*p)->ptr[0], key))
-	    return *p;
-	p++;
+    for (i = 0; i < RARRAY(ary)->len; ++i) {
+	v = RARRAY(ary)->ptr[i];
+	if (TYPE(v) == T_ARRAY &&
+	    RARRAY(v)->len > 0 &&
+	    rb_equal(RARRAY(v)->ptr[0], key))
+	    return v;
     }
     return Qnil;
 }
@@ -2393,17 +2436,15 @@ VALUE
 rb_ary_rassoc(ary, value)
     VALUE ary, value;
 {
-    VALUE *p, *pend;
+    long i;
+    VALUE v;
 
-    p = RARRAY(ary)->ptr;
-    pend = p + RARRAY(ary)->len;
-
-    while (p < pend) {
-	if (TYPE(*p) == T_ARRAY
-	    && RARRAY(*p)->len > 1
-	    && rb_equal(RARRAY(*p)->ptr[1], value))
-	    return *p;
-	p++;
+    for (i = 0; i < RARRAY(ary)->len; ++i) {
+	v = RARRAY(ary)->ptr[i];
+	if (TYPE(v) == T_ARRAY &&
+	    RARRAY(v)->len > 1 &&
+	    rb_equal(RARRAY(v)->ptr[1], value))
+	    return v;
     }
     return Qnil;
 }
@@ -2437,7 +2478,7 @@ rb_ary_equal(ary1, ary2)
     }
     if (RARRAY(ary1)->len != RARRAY(ary2)->len) return Qfalse;
     for (i=0; i<RARRAY(ary1)->len; i++) {
-	if (!rb_equal(RARRAY(ary1)->ptr[i], RARRAY(ary2)->ptr[i]))
+	if (!rb_equal(rb_ary_elt(ary1, i), rb_ary_elt(ary2, i)))
 	    return Qfalse;
     }
     return Qtrue;
@@ -2461,7 +2502,7 @@ rb_ary_eql(ary1, ary2)
     if (TYPE(ary2) != T_ARRAY) return Qfalse;
     if (RARRAY(ary1)->len != RARRAY(ary2)->len) return Qfalse;
     for (i=0; i<RARRAY(ary1)->len; i++) {
-	if (!rb_eql(RARRAY(ary1)->ptr[i], RARRAY(ary2)->ptr[i]))
+	if (!rb_eql(rb_ary_elt(ary1, i), rb_ary_elt(ary2, i)))
 	    return Qfalse;
     }
     return Qtrue;
@@ -2552,7 +2593,7 @@ rb_ary_cmp(ary1, ary2)
 	len = RARRAY(ary2)->len;
     }
     for (i=0; i<len; i++) {
-	VALUE v = rb_funcall(RARRAY(ary1)->ptr[i], id_cmp, 1, RARRAY(ary2)->ptr[i]);
+	VALUE v = rb_funcall(rb_ary_elt(ary1, i), id_cmp, 1, rb_ary_elt(ary2, i));
 	if (v != INT2FIX(0)) {
 	    return v;
 	}
@@ -2605,7 +2646,7 @@ rb_ary_diff(ary1, ary2)
 
     for (i=0; i<RARRAY(ary1)->len; i++) {
 	if (st_lookup(RHASH(hash)->tbl, RARRAY(ary1)->ptr[i], 0)) continue;
-	rb_ary_push(ary3, RARRAY(ary1)->ptr[i]);
+	rb_ary_push(ary3, rb_ary_elt(ary1, i));
     }
     return ary3;
 }
@@ -2625,7 +2666,7 @@ static VALUE
 rb_ary_and(ary1, ary2)
     VALUE ary1, ary2;
 {
-    VALUE hash, ary3;
+    VALUE hash, ary3, v, vv;
     long i;
 
     ary2 = to_ary(ary2);
@@ -2634,9 +2675,9 @@ rb_ary_and(ary1, ary2)
     hash = ary_make_hash(ary2, 0);
 
     for (i=0; i<RARRAY(ary1)->len; i++) {
-	VALUE v = RARRAY(ary1)->ptr[i];
-	if (st_delete(RHASH(hash)->tbl, (st_data_t*)&v, 0)) {
-	    rb_ary_push(ary3, RARRAY(ary1)->ptr[i]);
+	v = vv = rb_ary_elt(ary1, i);
+	if (st_delete(RHASH(hash)->tbl, (st_data_t*)&vv, 0)) {
+	    rb_ary_push(ary3, v);
 	}
     }
 
@@ -2659,7 +2700,7 @@ rb_ary_or(ary1, ary2)
     VALUE ary1, ary2;
 {
     VALUE hash, ary3;
-    VALUE v;
+    VALUE v, vv;
     long i;
 
     ary2 = to_ary(ary2);
@@ -2667,15 +2708,15 @@ rb_ary_or(ary1, ary2)
     hash = ary_make_hash(ary1, ary2);
 
     for (i=0; i<RARRAY(ary1)->len; i++) {
-	v = RARRAY(ary1)->ptr[i];
-	if (st_delete(RHASH(hash)->tbl, (st_data_t*)&v, 0)) {
-	    rb_ary_push(ary3, RARRAY(ary1)->ptr[i]);
+	v = vv = rb_ary_elt(ary1, i);
+	if (st_delete(RHASH(hash)->tbl, (st_data_t*)&vv, 0)) {
+	    rb_ary_push(ary3, v);
 	}
     }
     for (i=0; i<RARRAY(ary2)->len; i++) {
-	v = RARRAY(ary2)->ptr[i];
-	if (st_delete(RHASH(hash)->tbl, (st_data_t*)&v, 0)) {
-	    rb_ary_push(ary3, RARRAY(ary2)->ptr[i]);
+	v = vv = rb_ary_elt(ary2, i);
+	if (st_delete(RHASH(hash)->tbl, (st_data_t*)&vv, 0)) {
+	    rb_ary_push(ary3, v);
 	}
     }
     return ary3;
@@ -2699,26 +2740,21 @@ static VALUE
 rb_ary_uniq_bang(ary)
     VALUE ary;
 {
-    VALUE hash;
-    VALUE *p, *q, *end;
-
-    rb_ary_modify(ary); 
+    VALUE hash, v, vv;
+    long i, j;
 
     hash = ary_make_hash(ary, 0);
 
     if (RARRAY(ary)->len == RHASH(hash)->tbl->num_entries) {
 	return Qnil;
     }
-    p = q = RARRAY(ary)->ptr;
-    end = p + RARRAY(ary)->len;
-    while (p < end) {
-	VALUE v = *p;
-	if (st_delete(RHASH(hash)->tbl, (st_data_t*)&v, 0)) {
-	    *q++ = *p;
+    for (i=j=0; i<RARRAY(ary)->len; i++) {
+	v = vv = rb_ary_elt(ary, i);
+	if (st_delete(RHASH(hash)->tbl, (st_data_t*)&vv, 0)) {
+	    rb_ary_store(ary, j++, v);
 	}
-	p++;
     }
-    RARRAY(ary)->len = (q - RARRAY(ary)->ptr);
+    RARRAY(ary)->len = j;
 
     return ary;
 }
@@ -2837,10 +2873,13 @@ flatten(ary, idx, ary2, memo)
 	rb_raise(rb_eArgError, "tried to flatten recursive array");
     }
     rb_ary_push(memo, id);
-    rb_ary_update(ary, idx, 1, ary2);
+    rb_ary_splice(ary, idx, 1, ary2);
     while (i < lim) {
-	if (TYPE(RARRAY(ary)->ptr[i]) == T_ARRAY) {
-	    n = flatten(ary, i, RARRAY(ary)->ptr[i], memo);
+	VALUE tmp;
+
+	tmp = rb_check_array_type(rb_ary_elt(ary, i));
+	if (!NIL_P(tmp)) {
+	    n = flatten(ary, i, tmp, memo);
 	    i += n; lim += n;
 	}
 	i++;
@@ -2872,15 +2911,16 @@ rb_ary_flatten_bang(ary)
     int mod = 0;
     VALUE memo = Qnil;
 
-    rb_ary_modify(ary);
     while (i<RARRAY(ary)->len) {
 	VALUE ary2 = RARRAY(ary)->ptr[i];
+	VALUE tmp;
 
-	if (TYPE(ary2) == T_ARRAY) {
+	tmp = rb_check_array_type(ary2);
+	if (!NIL_P(tmp)) {
 	    if (NIL_P(memo)) {
 		memo = rb_ary_new();
 	    }
-	    i += flatten(ary, i, ary2, memo);
+	    i += flatten(ary, i, tmp, memo);
 	    mod = 1;
 	}
 	i++;
@@ -2973,7 +3013,7 @@ Init_Array()
     rb_define_method(rb_cArray, "collect!", rb_ary_collect_bang, 0);
     rb_define_method(rb_cArray, "map", rb_ary_collect, 0);
     rb_define_method(rb_cArray, "map!", rb_ary_collect_bang, 0);
-    rb_define_method(rb_cArray, "select", rb_ary_select, -1);
+    rb_define_method(rb_cArray, "select", rb_ary_select, 0);
     rb_define_method(rb_cArray, "values_at", rb_ary_values_at, -1);
     rb_define_method(rb_cArray, "delete", rb_ary_delete, 1);
     rb_define_method(rb_cArray, "delete_at", rb_ary_delete_at_m, 1);
