@@ -2,8 +2,8 @@
 /*
  * rubyext.c
  *
- * $Author: why $
- * $Date: 2005/09/13 03:58:31 $
+ * $Author: matz $
+ * $Date: 2005/10/30 18:20:52 $
  *
  * Copyright (C) 2003-2005 why the lucky stiff
  */
@@ -46,13 +46,6 @@ typedef struct {
 
 #define RUBY_DOMAIN   "ruby.yaml.org,2002"
 
-#ifndef StringValue
-#define StringValue(v) (v)
-#endif
-#ifndef rb_attr_get
-#define rb_attr_get(o, i)  rb_ivar_get(o, i)
-#endif
-
 /*
  * symbols and constants
  */
@@ -83,6 +76,9 @@ SyckNode * rb_syck_bad_anchor_handler _((SyckParser *, char *));
 void rb_syck_output_handler _((SyckEmitter *, char *, long));
 void rb_syck_emitter_handler _((SyckEmitter *, st_data_t));
 int syck_parser_assign_io _((SyckParser *, VALUE));
+VALUE syck_scalar_alloc _((VALUE class));
+VALUE syck_seq_alloc _((VALUE class));
+VALUE syck_map_alloc _((VALUE class));
 
 struct parser_xtra {
     VALUE data;  /* Borrowed this idea from marshal.c to fix [ruby-core:8067] problem */
@@ -774,6 +770,7 @@ syck_parser_initialize(argc, argv, self)
         Check_Type(options, T_HASH);
     }
     rb_ivar_set(self, s_options, options);
+    rb_ivar_set(self, s_input, Qnil);
     return self;
 }
 
@@ -1120,8 +1117,6 @@ syck_resolver_transfer( self, type, val )
                         }
                         else if ( rb_cObject == target_class && subclass_v == Qnil )
                         {
-                            // StringValue(subclass);
-                            // printf( "No class: %s\n", RSTRING(subclass)->ptr );
                             target_class = cYObject;
                             type = subclass;
                             subclass = cYObject;
@@ -1345,7 +1340,7 @@ syck_badalias_cmp( alias1, alias2 )
  */
 VALUE
 syck_domaintype_initialize( self, domain, type_id, val )
-    VALUE self, type_id, val;
+    VALUE self, domain, type_id, val;
 {
     rb_iv_set( self, "@domain", domain );
     rb_iv_set( self, "@type_id", type_id );
@@ -2028,6 +2023,7 @@ syck_emitter_reset( argc, argv, self )
         rb_ivar_set(self, s_options, options);
     }
     
+    emitter->headless = 1;
     emitter->bonus = (void *)bonus;
     rb_ivar_set(self, s_level, INT2FIX(0));
     rb_ivar_set(self, s_resolver, Qnil);
