@@ -2,8 +2,8 @@
 
   file.c -
 
-  $Author: matz $
-  $Date: 2004/07/17 02:23:51 $
+  $Author: ocean $
+  $Date: 2005/11/23 16:30:53 $
   created at: Mon Nov 15 12:24:34 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -64,17 +64,48 @@ char *strrchr _((const char*,const char));
 #include <sys/mkdev.h>
 #endif
 
-#ifndef HAVE_LSTAT
-#define lstat(path,st) stat(path,st)
+#if !defined HAVE_LSTAT && !defined lstat
+#define lstat stat
 #endif
+#if !HAVE_FSEEKO && !defined(fseeko)
+# define fseeko  fseek
+#endif
+
+#ifdef __BEOS__ /* should not change ID if -1 */
+static int
+be_chown(const char *path, uid_t owner, gid_t group)
+{
+    if (owner == -1 || group == -1) {
+	struct stat st;
+	if (stat(path, &st) < 0) return -1;
+	if (owner == -1) owner = st.st_uid;
+	if (group == -1) group = st.st_gid;
+    }
+    return chown(path, owner, group);
+}
+#define chown be_chown
+static int
+be_fchown(int fd, uid_t owner, gid_t group)
+{
+    if (owner == -1 || group == -1) {
+	struct stat st;
+	if (fstat(fd, &st) < 0) return -1;
+	if (owner == -1) owner = st.st_uid;
+	if (group == -1) group = st.st_gid;
+    }
+    return fchown(fd, owner, group);
+}
+#define fchown be_fchown
+#endif /* __BEOS__ */
 
 VALUE rb_cFile;
 VALUE rb_mFileTest;
 static VALUE rb_cStat;
 
+static long apply2files _((void (*)(const char *, void *), VALUE, void *));
 static long
 apply2files(func, vargs, arg)
-    void (*func)();
+    void (*func)_((const char *, void *));
     VALUE vargs;
     void *arg;
 {
@@ -93,7 +124,7 @@ apply2files(func, vargs, arg)
 
 /*
  *  call-seq:
- *     file.path -> file_ame
+ *     file.path -> filename
  *  
  *  Returns the pathname used to create <i>file</i> as a string. Does
  *  not normalize the name.
@@ -110,6 +141,7 @@ rb_file_path(obj)
     OpenFile *fptr;
 
     fptr = RFILE(rb_io_taint_check(obj))->fptr;
+    rb_io_check_initialized(fptr);
     if (!fptr->path) return Qnil;
     return rb_tainted_str_new2(fptr->path);
 }
@@ -175,6 +207,20 @@ rb_stat_cmp(self, other)
     return Qnil;
 }
 
+static VALUE rb_stat_dev _((VALUE));
+static VALUE rb_stat_ino _((VALUE));
+static VALUE rb_stat_mode _((VALUE));
+static VALUE rb_stat_nlink _((VALUE));
+static VALUE rb_stat_uid _((VALUE));
+static VALUE rb_stat_gid _((VALUE));
+static VALUE rb_stat_rdev _((VALUE));
+static VALUE rb_stat_size _((VALUE));
+static VALUE rb_stat_blksize _((VALUE));
+static VALUE rb_stat_blocks _((VALUE));
+static VALUE rb_stat_atime _((VALUE));
+static VALUE rb_stat_mtime _((VALUE));
+static VALUE rb_stat_ctime _((VALUE));
+
 /*
  *  call-seq:
  *     stat.dev    => fixnum
@@ -543,23 +589,23 @@ rb_stat_inspect(self)
 {
     VALUE str;
     int i;
-    static struct {
-        char *name;
-        VALUE (*func)();
+    static const struct {
+	const char *name;
+	VALUE (*func)_((VALUE));
     } member[] = {
-        {"dev",     rb_stat_dev},
-        {"ino",     rb_stat_ino},
-        {"mode",    rb_stat_mode},
-        {"nlink",   rb_stat_nlink},
-        {"uid",     rb_stat_uid},
-        {"gid",     rb_stat_gid},
-        {"rdev",    rb_stat_rdev},
-        {"size",    rb_stat_size},
-        {"blksize", rb_stat_blksize},
-        {"blocks",  rb_stat_blocks},
-        {"atime",   rb_stat_atime},
-        {"mtime",   rb_stat_mtime},
-        {"ctime",   rb_stat_ctime},
+	{"dev",	    rb_stat_dev},
+	{"ino",	    rb_stat_ino},
+	{"mode",    rb_stat_mode},
+	{"nlink",   rb_stat_nlink},
+	{"uid",	    rb_stat_uid},
+	{"gid",	    rb_stat_gid},
+	{"rdev",    rb_stat_rdev},
+	{"size",    rb_stat_size},
+	{"blksize", rb_stat_blksize},
+	{"blocks",  rb_stat_blocks},
+	{"atime",   rb_stat_atime},
+	{"mtime",   rb_stat_mtime},
+	{"ctime",   rb_stat_ctime},
     };
 
     str = rb_str_buf_new2("#<");
@@ -616,6 +662,43 @@ rb_stat(file, st)
     return stat(StringValueCStr(file), st);
 }
 
+#ifdef _WIN32
+static BOOL
+w32_io_info(file, st)
+    VALUE *file;
+    BY_HANDLE_FILE_INFORMATION *st;
+{
+    VALUE tmp;
+    HANDLE f;
+    BOOL ret = FALSE;
+
+    tmp = rb_check_convert_type(*file, T_FILE, "IO", "to_io");
+    if (!NIL_P(tmp)) {
+	OpenFile *fptr;
+
+	*file = Qnil;
+	GetOpenFile(tmp, fptr);
+	f = (HANDLE)rb_w32_get_osfhandle(fileno(fptr->f));
+	if (f == (HANDLE)-1) return FALSE;
+    }
+    else {
+	SafeStringValue(*file);
+	f = CreateFile(StringValueCStr(*file), 0,
+	               FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
+	               rb_w32_iswin95() ? 0 : FILE_FLAG_BACKUP_SEMANTICS, NULL);
+	if (f == INVALID_HANDLE_VALUE) return FALSE;
+    }
+    if (GetFileType(f) == FILE_TYPE_DISK) {
+	ZeroMemory(st, sizeof(*st));
+	ret = GetFileInformationByHandle(f, st);
+    }
+    if (NIL_P(tmp)) {
+	CloseHandle(f);
+    }
+    return ret;
+}
+#endif
+
 /*
  *  call-seq:
  *     File.stat(file_name)   =>  stat
@@ -1263,7 +1346,7 @@ check3rdbyte(fname, mode)
  * call-seq:
  *   File.setuid?(file_name)   =>  true or false
  *
- * Returns <code>true</code> if the named file is a has the setuid bit set.
+ * Returns <code>true</code> if the named file has the setuid bit set.
  */
 
 static VALUE
@@ -1281,7 +1364,7 @@ test_suid(obj, fname)
  * call-seq:
  *   File.setgid?(file_name)   =>  true or false
  *
- * Returns <code>true</code> if the named file is a has the setgid bit set.
+ * Returns <code>true</code> if the named file has the setgid bit set.
  */
 
 static VALUE
@@ -1299,7 +1382,7 @@ test_sgid(obj, fname)
  * call-seq:
  *   File.sticky?(file_name)   =>  true or false
  *
- * Returns <code>true</code> if the named file is a has the sticky bit set.
+ * Returns <code>true</code> if the named file has the sticky bit set.
  */
 
 static VALUE
@@ -1313,6 +1396,56 @@ test_sticky(obj, fname)
 #endif
 }
 
+/*
+ * call-seq:
+ *   File.identical?(file_1, file_2)   =>  true or false
+ *
+ * Returns <code>true</code> if the named files are identical.
+ */
+
+static VALUE
+test_identical(obj, fname1, fname2)
+    VALUE obj, fname1, fname2;
+{
+#ifndef DOSISH
+    struct stat st1, st2;
+
+    if (rb_stat(fname1, &st1) < 0) return Qfalse;
+    if (rb_stat(fname2, &st2) < 0) return Qfalse;
+    if (st1.st_dev != st2.st_dev) return Qfalse;
+    if (st1.st_ino != st2.st_ino) return Qfalse;
+#else
+#ifdef _WIN32
+    BY_HANDLE_FILE_INFORMATION st1, st2;
+#endif
+
+    rb_secure(2);
+#ifdef _WIN32
+    if (!w32_io_info(&fname1, &st1)) return Qfalse;
+    if (!w32_io_info(&fname2, &st2)) return Qfalse;
+
+    if (st1.dwVolumeSerialNumber == st2.dwVolumeSerialNumber &&
+	st1.nFileIndexHigh == st2.nFileIndexHigh &&
+	st1.nFileIndexLow == st2.nFileIndexLow)
+	return Qtrue;
+    if (NIL_P(fname1) || NIL_P(fname2)) return Qfalse;
+    if (rb_w32_iswin95()) return Qfalse;
+#else
+    SafeStringValue(fname1);
+    fname1 = rb_str_new4(fname1);
+    SafeStringValue(fname2);
+    if (access(RSTRING(fname1)->ptr, 0)) return Qfalse;
+    if (access(RSTRING(fname2)->ptr, 0)) return Qfalse;
+#endif
+    fname1 = rb_file_expand_path(fname1, Qnil);
+    fname2 = rb_file_expand_path(fname2, Qnil);
+    if (RSTRING(fname1)->len != RSTRING(fname2)->len) return Qfalse;
+    if (rb_memcicmp(RSTRING(fname1)->ptr, RSTRING(fname2)->ptr, RSTRING(fname1)->len))
+	return Qfalse;
+#endif
+    return Qtrue;
+}
+
 /*
  * call-seq:
  *    File.size(file_name)   => integer
@@ -1541,12 +1674,13 @@ rb_file_ctime(obj)
     return rb_time_new(st.st_ctime, 0);
 }
 
+static void chmod_internal _((const char *, void *));
 static void
 chmod_internal(path, mode)
     const char *path;
-    int mode;
+    void *mode;
 {
-    if (chmod(path, mode) < 0)
+    if (chmod(path, (int)mode) < 0)
 	rb_sys_fail(path);
 }
 
@@ -1618,12 +1752,13 @@ rb_file_chmod(obj, vmode)
 }
 
 #if defined(HAVE_LCHMOD)
+static void lchmod_internal _((const char *, void *));
 static void
 lchmod_internal(path, mode)
     const char *path;
-    int mode;
+    void *mode;
 {
-    if (lchmod(path, mode) < 0)
+    if (lchmod(path, (int)mode) < 0)
 	rb_sys_fail(path);
 }
 
@@ -1644,8 +1779,7 @@ rb_file_s_lchmod(argc, argv)
 {
     VALUE vmode;
     VALUE rest;
-    int mode;
-    long n;
+    long mode, n;
 
     rb_secure(2);
     rb_scan_args(argc, argv, "1*", &vmode, &rest);
@@ -1669,11 +1803,13 @@ struct chown_args {
     int owner, group;
 };
 
+static void chown_internal _((const char *, void *));
 static void
-chown_internal(path, args)
+chown_internal(path, argp)
     const char *path;
-    struct chown_args *args;
+    void *argp;
 {
+    struct chown_args *args = (struct chown_args *)argp;
     if (chown(path, args->owner, args->group) < 0)
 	rb_sys_fail(path);
 }
@@ -1741,15 +1877,18 @@ rb_file_chown(obj, owner, group)
     VALUE obj, owner, group;
 {
     OpenFile *fptr;
+    int o, g;
 
     rb_secure(2);
+    o = NIL_P(owner) ? -1 : NUM2INT(owner);
+    g = NIL_P(group) ? -1 : NUM2INT(group);
     GetOpenFile(obj, fptr);
 #if defined(DJGPP) || defined(__CYGWIN32__) || defined(_WIN32) || defined(__EMX__)
     if (!fptr->path) return Qnil;
-    if (chown(fptr->path, NUM2INT(owner), NUM2INT(group)) == -1)
+    if (chown(fptr->path, o, g) == -1)
 	rb_sys_fail(fptr->path);
 #else
-    if (fchown(fileno(fptr->f), NUM2INT(owner), NUM2INT(group)) == -1)
+    if (fchown(fileno(fptr->f), o, g) == -1)
 	rb_sys_fail(fptr->path);
 #endif
 
@@ -1757,11 +1896,13 @@ rb_file_chown(obj, owner, group)
 }
 
 #if defined(HAVE_LCHOWN) && !defined(__CHECKER__)
+static void lchown_internal _((const char *, void *));
 static void
-lchown_internal(path, args)
+lchown_internal(path, argp)
     const char *path;
-    struct chown_args *args;
+    void *argp;
 {
+    struct chown_args *args = (struct chown_args *)argp;
     if (lchown(path, args->owner, args->group) < 0)
 	rb_sys_fail(path);
 }
@@ -1817,13 +1958,16 @@ rb_file_s_lchown(argc, argv)
 
 struct timeval rb_time_timeval();
 
+static void utime_internal _((const char *, void *));
+
 #if defined(HAVE_UTIMES) && !defined(__CHECKER__)
 
 static void
-utime_internal(path, tvp)
-    char *path;
-    struct timeval tvp[];
+utime_internal(path, arg)
+    const char *path;
+    void *arg;
 {
+    struct timeval *tvp = arg;
     if (utimes(path, tvp) < 0)
 	rb_sys_fail(path);
 }
@@ -1865,10 +2009,11 @@ struct utimbuf {
 #endif
 
 static void
-utime_internal(path, utp)
+utime_internal(path, arg)
     const char *path;
-    struct utimbuf *utp;
+    void *arg;
 {
+    struct utimbuf *utp = arg;
     if (utime(path, utp) < 0)
 	rb_sys_fail(path);
 }
@@ -1926,6 +2071,7 @@ static VALUE
 rb_file_s_link(klass, from, to)
     VALUE klass, from, to;
 {
+#ifdef HAVE_LINK
     SafeStringValue(from);
     SafeStringValue(to);
 
@@ -1933,6 +2079,10 @@ rb_file_s_link(klass, from, to)
 	sys_fail2(from, to);
     }
     return INT2FIX(0);
+#else
+    rb_notimplement();
+    return Qnil;		/* not reached */
+#endif
 }
 
 /*
@@ -1988,7 +2138,11 @@ rb_file_s_readlink(klass, path)
 
     SafeStringValue(path);
     buf = xmalloc(size);
-    while ((rv = readlink(StringValueCStr(path), buf, size)) == size) {
+    while ((rv = readlink(RSTRING(path)->ptr, buf, size)) == size
+#ifdef _AIX
+	    || (rv < 0 && errno == ERANGE) /* quirky behavior of GPFS */
+#endif
+	) {
 	size *= 2;
 	buf = xrealloc(buf, size);
     }
@@ -2006,9 +2160,11 @@ rb_file_s_readlink(klass, path)
 #endif
 }
 
+static void unlink_internal _((const char *, void *));
 static void
-unlink_internal(path)
+unlink_internal(path, arg)
     const char *path;
+    void *arg;
 {
     if (unlink(path) < 0)
 	rb_sys_fail(path);
@@ -2197,7 +2353,11 @@ rb_path_next(s)
     return (char *)s;
 }
 
+#if defined(DOSISH_UNC) || defined(DOSISH_DRIVE_LETTER) 
 #define skipprefix rb_path_skip_prefix
+#else
+#define skipprefix(path) (path)
+#endif
 char *
 rb_path_skip_prefix(path)
     const char *path;
@@ -2205,7 +2365,9 @@ rb_path_skip_prefix(path)
 #if defined(DOSISH_UNC) || defined(DOSISH_DRIVE_LETTER) 
 #ifdef DOSISH_UNC
     if (isdirsep(path[0]) && isdirsep(path[1])) {
-	if (*(path = nextdirsep(path + 2)))
+	path += 2;
+	while (isdirsep(*path)) path++;
+	if (*(path = nextdirsep(path)) && path[1] && !isdirsep(path[1]))
 	    path = nextdirsep(path + 1);
 	return (char *)path;
     }
@@ -2560,27 +2722,46 @@ rb_file_s_basename(argc, argv)
     VALUE *argv;
 {
     VALUE fname, fext, basename;
-    char *name, *p, *ext = NULL;
+    char *name, *p;
+#if defined DOSISH_DRIVE_LETTER || defined DOSISH_UNC
+    char *root;
+#endif
     int f;
 
     if (rb_scan_args(argc, argv, "11", &fname, &fext) == 2) {
-	ext = StringValueCStr(fext);
+	StringValue(fext);
     }
     StringValue(fname);
     if (RSTRING(fname)->len == 0 || !*(name = RSTRING(fname)->ptr))
 	return fname;
-    if (!*(name = skiproot(name))) {
+    name = skipprefix(name);
+#if defined DOSISH_DRIVE_LETTER || defined DOSISH_UNC
+    root = name;
+#endif
+    while (isdirsep(*name))
+	name++;
+    if (!*name) {
 	p = name - 1;
 	f = 1;
+#if defined DOSISH_DRIVE_LETTER || defined DOSISH_UNC
+	if (name != root) {
+	    /* has slashes */
+	}
 #ifdef DOSISH_DRIVE_LETTER
-	if (*p == ':') {
+	else if (*p == ':') {
 	    p++;
 	    f = 0;
 	}
+#endif
+#ifdef DOSISH_UNC
+	else {
+	    p = "/";
+	}
+#endif
 #endif
     }
     else if (!(p = strrdirsep(name))) {
-	if (NIL_P(fext) || !(f = rmext(name, ext))) {
+	if (NIL_P(fext) || !(f = rmext(name, StringValueCStr(fext)))) {
 	    f = chompdirsep(name) - name;
 	    if (f == RSTRING(fname)->len) return fname;
 	}
@@ -2588,7 +2769,7 @@ rb_file_s_basename(argc, argv)
     }
     else {
 	while (isdirsep(*p)) p++; /* skip last / */
-	if (NIL_P(fext) || !(f = rmext(p, ext))) {
+	if (NIL_P(fext) || !(f = rmext(p, StringValueCStr(fext)))) {
 	    f = chompdirsep(p) - p;
 	}
     }
@@ -2619,8 +2800,8 @@ rb_file_s_dirname(klass, fname)
     name = StringValueCStr(fname);
     root = skiproot(name);
 #ifdef DOSISH_UNC
-    if (root > name + 2 && isdirsep(*name))
-	name = root - 2;
+    if (root > name + 1 && isdirsep(*name))
+	root = skipprefix(name = root - 2);
 #else
     if (root > name + 1)
 	name = root - 1;
@@ -2631,9 +2812,16 @@ rb_file_s_dirname(klass, fname)
     }
     if (p == name)
 	return rb_str_new2(".");
+#ifdef DOSISH_DRIVE_LETTER
+    if (has_drive_letter(name) && isdirsep(*(name + 2))) {
+	dirname = rb_str_new(name, 3);
+	rb_str_cat(dirname, skiproot(name + 2), p - skiproot(name + 2));
+    }
+    else
+#endif
     dirname = rb_str_new(name, p - name);
 #ifdef DOSISH_DRIVE_LETTER
-    if (root == name + 2 && name[1] == ':')
+    if (has_drive_letter(name) && root == name + 2 && p - name == 2)
 	rb_str_cat(dirname, ".", 1);
 #endif
     OBJ_INFECT(dirname, fname);
@@ -2669,7 +2857,7 @@ rb_file_s_extname(klass, fname)
  	p++;
  
      e = strrchr(p, '.');	/* get the last dot of the last component */
-     if (!e || e == p)		/* no dot, or the only dot is first? */
+     if (!e || e == p || !e[1])	/* no dot, or the only dot is first or end? */
 	 return rb_str_new2("");
      extname = rb_str_new(e, chompdirsep(e) - e);	/* keep the dot, too! */
      OBJ_INFECT(extname, fname);
@@ -2714,7 +2902,7 @@ rb_file_join(ary, sep)
     long len, i;
     int taint = 0;
     VALUE result, tmp;
-    char *name;
+    char *name, *tail;
 
     if (RARRAY(ary)->len == 0) return rb_str_new(0, 0);
     if (OBJ_TAINTED(ary)) taint = 1;
@@ -2751,11 +2939,18 @@ rb_file_join(ary, sep)
 	    }
 	    break;
 	  default:
-	    tmp = rb_obj_as_string(tmp);
+	    StringValueCStr(tmp);
 	}
 	name = StringValueCStr(result);
-	if (i > 0 && !NIL_P(sep) && !*chompdirsep(name))
-	    rb_str_buf_append(result, sep);
+	if (i > 0 && !NIL_P(sep)) {
+	    tail = chompdirsep(name);
+	    if (RSTRING(tmp)->ptr && isdirsep(RSTRING(tmp)->ptr[0])) {
+		RSTRING(result)->len = tail - name;
+	    }
+	    else if (!*tail) {
+		rb_str_buf_append(result, sep);
+	    }
+	}
 	rb_str_buf_append(result, tmp);
 	if (OBJ_TAINTED(tmp)) taint = 1;
     }
@@ -2801,11 +2996,14 @@ static VALUE
 rb_file_s_truncate(klass, path, len)
     VALUE klass, path, len;
 {
+    off_t pos;
+
     rb_secure(2);
+    pos = NUM2OFFT(len);
     SafeStringValue(path);
 
 #ifdef HAVE_TRUNCATE
-    if (truncate(StringValueCStr(path), NUM2OFFT(len)) < 0)
+    if (truncate(StringValueCStr(path), pos) < 0)
 	rb_sys_fail(RSTRING(path)->ptr);
 #else
 # ifdef HAVE_CHSIZE
@@ -2821,7 +3019,7 @@ rb_file_s_truncate(klass, path, len)
 	    rb_sys_fail(RSTRING(path)->ptr);
 	}
 #  endif
-	if (chsize(tmpfd, NUM2OFFT(len)) < 0) {
+	if (chsize(tmpfd, pos) < 0) {
 	    close(tmpfd);
 	    rb_sys_fail(RSTRING(path)->ptr);
 	}
@@ -2853,18 +3051,24 @@ rb_file_truncate(obj, len)
     VALUE obj, len;
 {
     OpenFile *fptr;
+    FILE *f;
+    off_t pos;
 
     rb_secure(2);
+    pos = NUM2OFFT(len);
     GetOpenFile(obj, fptr);
     if (!(fptr->mode & FMODE_WRITABLE)) {
 	rb_raise(rb_eIOError, "not opened for writing");
     }
+    f = GetWriteFile(fptr);
+    fflush(f);
+    fseeko(f, (off_t)0, SEEK_CUR);
 #ifdef HAVE_TRUNCATE
-    if (ftruncate(fileno(fptr->f), NUM2OFFT(len)) < 0)
+    if (ftruncate(fileno(f), pos) < 0)
 	rb_sys_fail(fptr->path);
 #else
 # ifdef HAVE_CHSIZE
-    if (chsize(fileno(fptr->f), NUM2OFFT(len)) < 0)
+    if (chsize(fileno(f), pos) < 0)
 	rb_sys_fail(fptr->path);
 # else
     rb_notimplement();
@@ -2886,26 +3090,30 @@ rb_file_truncate(obj, len)
 #  define LOCK_UN 8
 # endif
 
-#if 0
+#if 1
 static int
 rb_thread_flock(fd, op, fptr)
     int fd, op;
     OpenFile *fptr;
 {
     if (rb_thread_alone() || (op & LOCK_NB)) {
-	return flock(fd, op);
+	int ret;
+	TRAP_BEG;
+	ret = flock(fd, op);
+	TRAP_END;
+	return ret;
     }
     op |= LOCK_NB;
     while (flock(fd, op) < 0) {
 	switch (errno) {
-          case EAGAIN:
-          case EACCES:
+	  case EAGAIN:
+	  case EACCES:
 #if defined(EWOULDBLOCK) && EWOULDBLOCK != EAGAIN
 	  case EWOULDBLOCK:
 #endif
 	    rb_thread_polling();	/* busy wait */
 	    rb_io_check_closed(fptr);
-            continue;
+	    continue;
 	  default:
 	    return -1;
 	}
@@ -2951,28 +3159,30 @@ rb_file_flock(obj, operation)
 {
 #ifndef __CHECKER__
     OpenFile *fptr;
+    int op;
 
     rb_secure(2);
+    op = NUM2INT(operation);
     GetOpenFile(obj, fptr);
 
     if (fptr->mode & FMODE_WRITABLE) {
 	fflush(GetWriteFile(fptr));
     }
   retry:
-    if (flock(fileno(fptr->f), NUM2INT(operation)) < 0) {
-        switch (errno) {
-          case EAGAIN:
-          case EACCES:
+    if (flock(fileno(fptr->f), op) < 0) {
+	switch (errno) {
+	  case EAGAIN:
+	  case EACCES:
 #if defined(EWOULDBLOCK) && EWOULDBLOCK != EAGAIN
-          case EWOULDBLOCK:
+	  case EWOULDBLOCK:
 #endif
-              return Qfalse;
+	      return Qfalse;
 	  case EINTR:
 #if defined(ERESTART)
 	  case ERESTART:
 #endif
 	    goto retry;
-        }
+	}
 	rb_sys_fail(fptr->path);
     }
 #endif
@@ -3021,14 +3231,14 @@ test_check(n, argc, argv)
  *     ?d  | boolean | True if file1 exists and is a directory
  *     ?e  | boolean | True if file1 exists
  *     ?f  | boolean | True if file1 exists and is a regular file
- *     ?g  | boolean | True if files has the \CF{setgid} bit
+ *     ?g  | boolean | True if file1 has the \CF{setgid} bit
  *         |         | set (false under NT)
  *     ?G  | boolean | True if file1 exists and has a group
  *         |         | ownership equal to the caller's group
  *     ?k  | boolean | True if file1 exists and has the sticky bit set
- *     ?l  | boolean | True if files exists and is a symbolic link
+ *     ?l  | boolean | True if file1 exists and is a symbolic link
  *     ?M  | Time    | Last modification time for file1
- *     ?o  | boolean | True if files exists and is owned by 
+ *     ?o  | boolean | True if file1 exists and is owned by 
  *         |         | the caller's effective uid
  *     ?O  | boolean | True if file1 exists and is owned by
  *         |         | the caller's real uid
@@ -3037,7 +3247,7 @@ test_check(n, argc, argv)
  *         |         | uid/gid of the caller
  *     ?R  | boolean | True if file is readable by the real
  *         |         | uid/gid of the caller
- *     ?s  | int/nil | If files has nonzero size, return the size,
+ *     ?s  | int/nil | If file1 has nonzero size, return the size,
  *         |         | otherwise return nil
  *     ?S  | boolean | True if file1 exists and is a socket
  *     ?u  | boolean | True if file1 has the setuid bit set
@@ -3167,7 +3377,12 @@ rb_f_test(argc, argv)
 	}
     }
 
-    if (strchr("-=<>", cmd)) {
+    if (cmd == '-') {
+	CHECK(2);
+	return test_identical(0, argv[1], argv[2]);
+    }
+
+    if (strchr("=<>", cmd)) {
 	struct stat st1, st2;
 
 	CHECK(2);
@@ -3175,11 +3390,6 @@ rb_f_test(argc, argv)
 	if (rb_stat(argv[2], &st2) < 0) return Qfalse;
 
 	switch (cmd) {
-	  case '-':
-	    if (st1.st_dev == st2.st_dev && st1.st_ino == st2.st_ino)
-		return Qtrue;
-            return Qfalse;
-
 	  case '=':
 	    if (st1.st_mtime == st2.st_mtime) return Qtrue;
 	    return Qfalse;
@@ -3191,7 +3401,7 @@ rb_f_test(argc, argv)
 	  case '<':
 	    if (st1.st_mtime < st2.st_mtime) return Qtrue;
 	    return Qfalse;
-        }
+	}
     }
     /* unknown command */
     rb_raise(rb_eArgError, "unknown command ?%c", cmd);
@@ -3253,6 +3463,7 @@ rb_stat_init(obj, fname)
     return Qnil;
 }
 
+/* :nodoc: */
 static VALUE
 rb_stat_init_copy(copy, orig)
     VALUE copy, orig;
@@ -3822,11 +4033,12 @@ is_absolute_path(path)
 
 #ifndef DOSISH
 static int
-path_check_1(path)
-     VALUE path;
+path_check_0(fpath, loadpath)
+     VALUE fpath;
+     int loadpath;
 {
     struct stat st;
-    char *p0 = StringValueCStr(path);
+    char *p0 = StringValueCStr(fpath);
     char *p = 0, *s;
 
     if (!is_absolute_path(p0)) {
@@ -3838,7 +4050,7 @@ path_check_1(path)
 
 	rb_str_cat2(newpath, "/");
 	rb_str_cat2(newpath, p0);
-	return path_check_1(newpath);
+	return path_check_0(newpath, loadpath);
     }
     for (;;) {
 #ifndef S_IWOTH
@@ -3846,7 +4058,7 @@ path_check_1(path)
 #endif
 	if (stat(p0, &st) == 0 && S_ISDIR(st.st_mode) && (st.st_mode & S_IWOTH)
 #ifdef S_ISVTX
-	    && !(st.st_mode & S_ISVTX)
+	    && (loadpath || !(st.st_mode & S_ISVTX))
 #endif
 	    ) {
 	    rb_warn("Insecure world writable dir %s, mode 0%o", p0, st.st_mode);
@@ -3862,6 +4074,17 @@ path_check_1(path)
 }
 #endif
 
+static int
+fpath_check(path)
+    char *path;
+{
+#ifndef DOSISH
+    return path_check_0(rb_str_new2(path), Qfalse);
+#else
+    return 1;
+#endif
+}
+
 int
 rb_path_check(path)
     char *path;
@@ -3878,7 +4101,7 @@ rb_path_check(path)
     if (!p) p = pend;
 
     for (;;) {
-	if (!path_check_1(rb_str_new(p0, p - p0))) {
+	if (!path_check_0(rb_str_new(p0, p - p0), Qtrue)) {
 	    return 0;		/* not safe */
 	}
 	p0 = p + 1;
@@ -3930,6 +4153,7 @@ rb_find_file_ext(filep, ext)
 	if (rb_safe_level() >= 2 && OBJ_TAINTED(fname)) {
 	    rb_raise(rb_eSecurityError, "loading from unsafe file %s", f);
 	}
+	OBJ_FREEZE(fname);
 	f = StringValueCStr(fname);
 	*filep = fname;
     }
@@ -3938,6 +4162,7 @@ rb_find_file_ext(filep, ext)
 	for (i=0; ext[i]; i++) {
 	    fname = rb_str_dup(*filep);
 	    rb_str_cat2(fname, ext[i]);
+	    OBJ_FREEZE(fname);
 	    if (file_load_ok(StringValueCStr(fname))) {
 		*filep = fname;
 		return i+1;
@@ -3958,6 +4183,7 @@ rb_find_file_ext(filep, ext)
 	for (j=0; ext[j]; j++) {
 	    fname = rb_str_dup(*filep);
 	    rb_str_cat2(fname, ext[j]);
+	    OBJ_FREEZE(fname);
 	    found = dln_find_file(StringValueCStr(fname), path);
 	    if (found && file_load_ok(found)) {
 		*filep = fname;
@@ -3981,12 +4207,13 @@ rb_find_file(path)
 	if (rb_safe_level() >= 1 && OBJ_TAINTED(path)) {
 	    rb_raise(rb_eSecurityError, "loading from unsafe path %s", f);
 	}
+	OBJ_FREEZE(path);
 	f = StringValueCStr(path);
     }
 
 #if defined(__MACOS__) || defined(riscos)
     if (is_macos_native_path(f)) {
-	if (rb_safe_level() >= 1 && !rb_path_check(f)) {
+	if (rb_safe_level() >= 1 && !fpath_check(f)) {
 	    rb_raise(rb_eSecurityError, "loading from unsafe file %s", f);
 	}
 	if (file_load_ok(f)) return path;
@@ -3994,7 +4221,7 @@ rb_find_file(path)
 #endif
 
     if (is_absolute_path(f)) {
-	if (rb_safe_level() >= 1 && !rb_path_check(f)) {
+	if (rb_safe_level() >= 1 && !fpath_check(f)) {
 	    rb_raise(rb_eSecurityError, "loading from unsafe file %s", f);
 	}
 	if (file_load_ok(f)) return path;
@@ -4035,11 +4262,13 @@ rb_find_file(path)
 	return 0;		/* no path, no load */
     }
     f = dln_find_file(f, lpath);
-    if (rb_safe_level() >= 1 && !rb_path_check(f)) {
+    if (rb_safe_level() >= 1 && !fpath_check(f)) {
 	rb_raise(rb_eSecurityError, "loading from unsafe file %s", f);
     }
     if (file_load_ok(f)) {
-	return rb_str_new2(f);
+	tmp = rb_str_new2(f);
+	OBJ_FREEZE(tmp);
+	return tmp;
     }
     return 0;
 }
@@ -4120,6 +4349,8 @@ Init_File()
     define_filetest_function("setgid?", test_sgid, 1);
     define_filetest_function("sticky?", test_sticky, 1);
 
+    define_filetest_function("identical?", test_identical, 2);
+
     rb_define_singleton_method(rb_cFile, "stat",  rb_file_s_stat, 1);
     rb_define_singleton_method(rb_cFile, "lstat", rb_file_s_lstat, 1);
     rb_define_singleton_method(rb_cFile, "ftype", rb_file_s_ftype, 1);
