@@ -2,8 +2,8 @@
 
   gc.c -
 
-  $Author: matz $
-  $Date: 2004/07/27 16:07:57 $
+  $Author: nobu $
+  $Date: 2005/10/18 17:36:20 $
   created at: Tue Oct  5 09:44:46 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -44,6 +44,10 @@ extern unsigned long __libc_ia64_register_backing_store_base;
 #endif
 #endif
 
+#if defined _WIN32 || defined __CYGWIN__
+#include <windows.h>
+#endif
+
 void re_free_registers _((struct re_registers*));
 void rb_io_fptr_finalize _((struct OpenFile*));
 
@@ -84,6 +88,7 @@ static unsigned long malloc_increase = 0;
 static unsigned long malloc_limit = GC_MALLOC_LIMIT;
 static void run_final();
 static VALUE nomem_error;
+static void garbage_collect();
 
 void
 rb_memerror()
@@ -111,11 +116,11 @@ ruby_xmalloc(size)
     malloc_increase += size;
 
     if (malloc_increase > malloc_limit) {
-	rb_gc();
+	garbage_collect();
     }
     RUBY_CRITICAL(mem = malloc(size));
     if (!mem) {
-	rb_gc();
+	garbage_collect();
 	RUBY_CRITICAL(mem = malloc(size));
 	if (!mem) {
 	    rb_memerror();
@@ -152,7 +157,7 @@ ruby_xrealloc(ptr, size)
     malloc_increase += size;
     RUBY_CRITICAL(mem = realloc(ptr, size));
     if (!mem) {
-	rb_gc();
+	garbage_collect();
 	RUBY_CRITICAL(mem = realloc(ptr, size));
 	if (!mem) {
 	    rb_memerror();
@@ -180,14 +185,14 @@ static st_table *finalizer_table = 0;
 /*
  *  call-seq:
  *     GC.enable    => true or false
- *  
+ *
  *  Enables garbage collection, returning <code>true</code> if garbage
  *  collection was previously disabled.
- *     
+ *
  *     GC.disable   #=> false
  *     GC.enable    #=> true
  *     GC.enable    #=> false
- *     
+ *
  */
 
 VALUE
@@ -202,13 +207,13 @@ rb_gc_enable()
 /*
  *  call-seq:
  *     GC.disable    => true or false
- *  
+ *
  *  Disables garbage collection, returning <code>true</code> if garbage
  *  collection was already disabled.
- *     
+ *
  *     GC.disable   #=> false
  *     GC.disable   #=> true
- *     
+ *
  */
 
 VALUE
@@ -291,7 +296,7 @@ typedef struct RVALUE {
 	struct RFile   file;
 	struct RNode   node;
 	struct RMatch  match;
-	struct RVarmap varmap; 
+	struct RVarmap varmap;
 	struct SCOPE   scope;
     } as;
 #ifdef GC_DEBUG
@@ -373,7 +378,7 @@ rb_newobj()
 {
     VALUE obj;
 
-    if (!freelist) rb_gc();
+    if (!freelist) garbage_collect();
 
     obj = (VALUE)freelist;
     freelist = freelist->as.free.next;
@@ -393,7 +398,7 @@ rb_data_object_alloc(klass, datap, dmark, dfree)
     RUBY_DATA_FUNC dfree;
 {
     NEWOBJ(data, struct RData);
-    Check_Type(klass, T_CLASS);
+    if (klass) Check_Type(klass, T_CLASS);
     OBJSETUP(data, klass, T_DATA);
     data->data = datap;
     data->dfree = dfree;
@@ -405,28 +410,37 @@ rb_data_object_alloc(klass, datap, dmark, dfree)
 extern st_table *rb_class_tbl;
 VALUE *rb_gc_stack_start = 0;
 
+#ifdef DJGPP
+/* set stack size (http://www.delorie.com/djgpp/v2faq/faq15_9.html) */
+unsigned int _stklen = 0x180000; /* 1.5 kB */
+#endif
+
 #if defined(DJGPP) || defined(_WIN32_WCE)
 static unsigned int STACK_LEVEL_MAX = 65535;
-#else
-#ifdef __human68k__
-extern unsigned int _stacksize;
+#elif defined(__human68k__)
+unsigned int _stacksize = 262144;
 # define STACK_LEVEL_MAX (_stacksize - 4096)
 # undef HAVE_GETRLIMIT
-#else
-#ifdef HAVE_GETRLIMIT
+#elif defined(HAVE_GETRLIMIT)
 static unsigned int STACK_LEVEL_MAX = 655300;
 #else
 # define STACK_LEVEL_MAX 655300
 #endif
-#endif
-#endif
 
 #ifdef C_ALLOCA
 # define SET_STACK_END VALUE stack_end; alloca(0);
 # define STACK_END (&stack_end)
 #else
 # if defined(__GNUC__) && defined(USE_BUILTIN_FRAME_ADDRESS) && !defined(__ia64__)
-#  define  SET_STACK_END    VALUE *stack_end = __builtin_frame_address(0)
+#  if ( __GNUC__ == 3 && __GNUC_MINOR__ > 0 ) || __GNUC__ > 3
+__attribute__ ((noinline))
+#  endif
+static VALUE *
+stack_end_address(void)
+{
+    return (VALUE *)__builtin_frame_address(0);
+}
+#  define  SET_STACK_END    VALUE *stack_end = stack_end_address()
 # else
 #  define  SET_STACK_END    VALUE *stack_end = alloca(1)
 # endif
@@ -605,9 +619,11 @@ mark_locations_array(x, n)
     register VALUE *x;
     register long n;
 {
+    VALUE v;
     while (n--) {
-	if (is_pointer_to_heap((void *)*x)) {
-	    gc_mark(*x, 0);
+        v = *x;
+	if (is_pointer_to_heap((void *)v)) {
+	    gc_mark(v, 0);
 	}
 	x++;
     }
@@ -633,13 +649,13 @@ mark_entry(key, value, lev)
     return ST_CONTINUE;
 }
 
-void
+static void
 mark_tbl(tbl, lev)
     st_table *tbl;
     int lev;
 {
     if (!tbl) return;
-    st_foreach(tbl, mark_entry, lev+1);
+    st_foreach(tbl, mark_entry, lev);
 }
 
 void
@@ -660,13 +676,13 @@ mark_keyvalue(key, value, lev)
     return ST_CONTINUE;
 }
 
-void
+static void
 mark_hash(tbl, lev)
     st_table *tbl;
     int lev;
 {
     if (!tbl) return;
-    st_foreach(tbl, mark_keyvalue, lev+1);
+    st_foreach(tbl, mark_keyvalue, lev);
 }
 
 void
@@ -687,7 +703,7 @@ rb_gc_mark_maybe(obj)
 
 #define GC_LEVEL_MAX 250
 
-void
+static void
 gc_mark(ptr, lev)
     VALUE ptr;
     int lev;
@@ -697,10 +713,10 @@ gc_mark(ptr, lev)
     obj = RANY(ptr);
     if (rb_special_const_p(ptr)) return; /* special const not marked */
     if (obj->as.basic.flags == 0) return;       /* free cell */
-    if (obj->as.basic.flags & FL_MARK) return;  /* already marked */ 
+    if (obj->as.basic.flags & FL_MARK) return;  /* already marked */
     obj->as.basic.flags |= FL_MARK;
 
-    if (lev > GC_LEVEL_MAX) {
+    if (lev > GC_LEVEL_MAX || (lev == 0 && ruby_stack_check())) {
 	if (!mark_stack_overflow) {
 	    if (mark_stack_ptr - mark_stack < MARK_STACK_MAX) {
 		*mark_stack_ptr = ptr;
@@ -712,7 +728,7 @@ gc_mark(ptr, lev)
 	}
 	return;
     }
-    gc_mark_children(ptr, lev);
+    gc_mark_children(ptr, lev+1);
 }
 
 void
@@ -735,7 +751,7 @@ gc_mark_children(ptr, lev)
     obj = RANY(ptr);
     if (rb_special_const_p(ptr)) return; /* special const not marked */
     if (obj->as.basic.flags == 0) return;       /* free cell */
-    if (obj->as.basic.flags & FL_MARK) return;  /* already marked */ 
+    if (obj->as.basic.flags & FL_MARK) return;  /* already marked */
     obj->as.basic.flags |= FL_MARK;
 
   marking:
@@ -864,13 +880,11 @@ gc_mark_children(ptr, lev)
 	  case NODE_BLOCK_ARG:
 	  case NODE_POSTEXE:
 	    break;
-#ifdef C_ALLOCA
 	  case NODE_ALLOCA:
 	    mark_locations_array((VALUE*)obj->as.node.u1.value,
 				 obj->as.node.u3.cnt);
 	    ptr = (VALUE)obj->as.node.u2.node;
 	    goto again;
-#endif
 
 	  default:		/* unlisted NODE */
 	    if (is_pointer_to_heap(obj->as.node.u1.node)) {
@@ -982,12 +996,47 @@ gc_mark_children(ptr, lev)
 
 static void obj_free _((VALUE));
 
+static void
+finalize_list(p)
+    RVALUE *p;
+{
+    while (p) {
+	RVALUE *tmp = p->as.free.next;
+	run_final((VALUE)p);
+	if (!FL_TEST(p, FL_SINGLETON)) { /* not freeing page */
+	    p->as.free.flags = 0;
+	    p->as.free.next = freelist;
+	    freelist = p;
+	}
+	p = tmp;
+    }
+}
+
+static void
+free_unused_heaps()
+{
+    int i, j;
+
+    for (i = j = 1; j < heaps_used; i++) {
+	if (heaps[i].limit == 0) {
+	    free(heaps[i].slot);
+	    heaps_used--;
+	}
+	else {
+	    if (i != j) {
+		heaps[j] = heaps[i];
+	    }
+	    j++;
+	}
+    }
+}
+
 static void
 gc_sweep()
 {
     RVALUE *p, *pend, *final_list;
     int freed = 0;
-    int i, j;
+    int i;
     unsigned long live = 0;
 
     if (ruby_in_compile && ruby_parser_stack_on_heap()) {
@@ -1067,35 +1116,10 @@ gc_sweep()
 
     /* clear finalization list */
     if (final_list) {
-	RVALUE *tmp;
-
-	if (rb_prohibit_interrupt || ruby_in_compile) {
-	    deferred_final_list = final_list;
-	    return;
-	}
-
-	for (p = final_list; p; p = tmp) {
-	    tmp = p->as.free.next;
-	    run_final((VALUE)p);
-	    if (!FL_TEST(p, FL_SINGLETON)) { /* not freeing page */
-		p->as.free.flags = 0;
-		p->as.free.next = freelist;
-		freelist = p;
-	    }
-	}
-    }
-    for (i = j = 1; j < heaps_used; i++) {
-	if (heaps[i].limit == 0) {
-	    free(heaps[i].slot);
-	    heaps_used--;
-	}
-	else {
-	    if (i != j) {
-		heaps[j] = heaps[i];
-	    }
-	    j++;
-	}
+	deferred_final_list = final_list;
+	return;
     }
+    free_unused_heaps();
 }
 
 void
@@ -1204,11 +1228,9 @@ obj_free(obj)
 		RUBY_CRITICAL(free(RANY(obj)->as.node.u1.tbl));
 	    }
 	    break;
-#ifdef C_ALLOCA
 	  case NODE_ALLOCA:
 	    RUBY_CRITICAL(free(RANY(obj)->as.node.u1.node));
 	    break;
-#endif
 	}
 	return;			/* no need to free iv_tbl */
 
@@ -1230,8 +1252,8 @@ obj_free(obj)
 	break;
 
       default:
-	rb_bug("gc_sweep(): unknown data type 0x%lx(%ld)", obj,
-	       RANY(obj)->as.basic.flags & T_MASK);
+	rb_bug("gc_sweep(): unknown data type 0x%lx(0x%lx)",
+	       RANY(obj)->as.basic.flags & T_MASK, obj);
     }
 }
 
@@ -1281,8 +1303,8 @@ int rb_setjmp (rb_jmp_buf);
 #endif /* __human68k__ or DJGPP */
 #endif /* __GNUC__ */
 
-void
-rb_gc()
+static void
+garbage_collect()
 {
     struct gc_list *list;
     struct FRAME * volatile frame; /* gcc 2.7.2.3 -O2 bug??  */
@@ -1304,10 +1326,10 @@ rb_gc()
     during_gc++;
 
     init_mark_stack();
-    
+
     /* mark frame stack */
     for (frame = ruby_frame; frame; frame = frame->prev) {
-	rb_gc_mark_frame(frame); 
+	rb_gc_mark_frame(frame);
 	if (frame->tmp) {
 	    struct FRAME *tmp = frame->tmp;
 	    while (tmp) {
@@ -1374,7 +1396,7 @@ rb_gc()
     rb_mark_generic_ivar_tbl();
 
     rb_gc_mark_parser();
-    
+
     /* gc_mark objects whose marking are not completed*/
     while (!MARK_STACK_EMPTY){
 	if (mark_stack_overflow){
@@ -1387,6 +1409,13 @@ rb_gc()
     gc_sweep();
 }
 
+void
+rb_gc()
+{
+    garbage_collect();
+    rb_gc_finalize_deferred();
+}
+
 /*
  *  call-seq:
  *     GC.start                     => nil
@@ -1394,7 +1423,7 @@ rb_gc()
  *     ObjectSpace.garbage_collect  => nil
  *
  *  Initiates garbage collection, unless manually disabled.
- *     
+ *
  */
 
 VALUE
@@ -1404,13 +1433,29 @@ rb_gc_start()
     return Qnil;
 }
 
+void
+ruby_set_stack_size(size)
+    size_t size;
+{
+#ifndef STACK_LEVEL_MAX
+    STACK_LEVEL_MAX = size / sizeof(VALUE);
+#endif
+}
+
 void
 Init_stack(addr)
     VALUE *addr;
 {
-#if defined(__human68k__)
-    extern void *_SEND;
-    rb_gc_stack_start = _SEND;
+#if defined(_WIN32) || defined(__CYGWIN__)
+    MEMORY_BASIC_INFORMATION m;
+    memset(&m, 0, sizeof(m));
+    VirtualQuery(&m, &m, sizeof(m));
+    rb_gc_stack_start =
+	STACK_UPPER((VALUE *)&m, (VALUE *)m.BaseAddress,
+		    (VALUE *)((char *)m.BaseAddress + m.RegionSize) - 1);
+#elif defined(STACK_END_ADDRESS)
+    extern void *STACK_END_ADDRESS;
+    rb_gc_stack_start = STACK_END_ADDRESS;
 #else
     if (!addr) addr = (VALUE *)&addr;
     STACK_UPPER(&addr, addr, ++addr);
@@ -1434,14 +1479,6 @@ Init_stack(addr)
 	    STACK_LEVEL_MAX = (rlim.rlim_cur - space) / sizeof(VALUE);
 	}
     }
-#if defined(__ia64__) && (!defined(__GNUC__) || __GNUC__ < 2 || defined(__OPTIMIZE__))
-    /* ruby crashes on IA64 if compiled with optimizer on */
-    /* when if STACK_LEVEL_MAX is greater than this magic number */
-    /* I know this is a kludge.  I suspect optimizer bug */
-#define IA64_MAGIC_STACK_LIMIT 49152
-    if (STACK_LEVEL_MAX > IA64_MAGIC_STACK_LIMIT)
-	STACK_LEVEL_MAX = IA64_MAGIC_STACK_LIMIT;
-#endif
 #endif
 }
 
@@ -1452,29 +1489,29 @@ Init_stack(addr)
  *  The <code>ObjectSpace</code> module contains a number of routines
  *  that interact with the garbage collection facility and allow you to
  *  traverse all living objects with an iterator.
- *     
+ *
  *  <code>ObjectSpace</code> also provides support for object
  *  finalizers, procs that will be called when a specific object is
  *  about to be destroyed by garbage collection.
- *     
+ *
  *     include ObjectSpace
- *     
- *     
+ *
+ *
  *     a = "A"
  *     b = "B"
  *     c = "C"
- *     
- *     
+ *
+ *
  *     define_finalizer(a, proc {|id| puts "Finalizer one on #{id}" })
  *     define_finalizer(a, proc {|id| puts "Finalizer two on #{id}" })
  *     define_finalizer(b, proc {|id| puts "Finalizer three on #{id}" })
- *     
+ *
  *  <em>produces:</em>
- *     
+ *
  *     Finalizer three on 537763470
  *     Finalizer one on 537763480
  *     Finalizer two on 537763480
- *     
+ *
  */
 
 void
@@ -1556,7 +1593,7 @@ os_obj_of(of)
 /*
  *  call-seq:
  *     ObjectSpace.each_object([module]) {|obj| ... } => fixnum
- *  
+ *
  *  Calls the block once for each living, nonimmediate object in this
  *  Ruby process. If <i>module</i> is specified, calls the block
  *  for only those classes or modules that match (or are a subclass of)
@@ -1566,15 +1603,15 @@ os_obj_of(of)
  *  never returned. In the example below, <code>each_object</code>
  *  returns both the numbers we defined and several constants defined in
  *  the <code>Math</code> module.
- *     
+ *
  *     a = 102.7
  *     b = 95       # Won't be returned
  *     c = 12345678987654321
  *     count = ObjectSpace.each_object(Numeric) {|x| p x }
  *     puts "Total count: #{count}"
- *     
+ *
  *  <em>produces:</em>
- *     
+ *
  *     12345678987654321
  *     102.7
  *     2.71828182845905
@@ -1583,7 +1620,7 @@ os_obj_of(of)
  *     1.7976931348623157e+308
  *     2.2250738585072e-308
  *     Total count: 7
- *     
+ *
  */
 
 static VALUE
@@ -1659,9 +1696,9 @@ call_final(os, obj)
 /*
  *  call-seq:
  *     ObjectSpace.undefine_finalizer(obj)
- *  
+ *
  *  Removes all finalizers for <i>obj</i>.
- *     
+ *
  */
 
 static VALUE
@@ -1677,10 +1714,10 @@ undefine_final(os, obj)
 /*
  *  call-seq:
  *     ObjectSpace.define_finalizer(obj, aProc=proc())
- *  
- *  Adds <i>aProc</i> as a finalizer, to be called when <i>obj</i>
- *  is about to be destroyed.
- *     
+ *
+ *  Adds <i>aProc</i> as a finalizer, to be called after <i>obj</i>
+ *  was destroyed.
+ *
  */
 
 static VALUE
@@ -1744,19 +1781,22 @@ run_final(obj)
 {
     long i;
     int status, critical_save = rb_thread_critical;
-    VALUE args[3], table;
+    VALUE args[3], table, objid;
 
+    objid = rb_obj_id(obj);	/* make obj into id */
     rb_thread_critical = Qtrue;
-    args[1] = rb_ary_new3(1, rb_obj_id(obj)); /* make obj into id */
+    args[1] = 0;
     args[2] = (VALUE)ruby_safe_level;
     for (i=0; i<RARRAY(finalizers)->len; i++) {
 	args[0] = RARRAY(finalizers)->ptr[i];
+	if (!args[1]) args[1] = rb_ary_new3(1, objid);
 	rb_protect((VALUE(*)_((VALUE)))run_single_final, (VALUE)args, &status);
     }
     if (finalizer_table && st_delete(finalizer_table, (st_data_t*)&obj, &table)) {
 	for (i=0; i<RARRAY(table)->len; i++) {
 	    VALUE final = RARRAY(table)->ptr[i];
 	    args[0] = RARRAY(final)->ptr[1];
+	    if (!args[1]) args[1] = rb_ary_new3(1, objid);
 	    args[2] = FIX2INT(RARRAY(final)->ptr[0]);
 	    rb_protect((VALUE(*)_((VALUE)))run_single_final, (VALUE)args, &status);
 	}
@@ -1764,6 +1804,18 @@ run_final(obj)
     rb_thread_critical = critical_save;
 }
 
+void
+rb_gc_finalize_deferred()
+{
+    RVALUE *p = deferred_final_list;
+
+    deferred_final_list = 0;
+    if (p) {
+	finalize_list(p);
+	free_unused_heaps();
+    }
+}
+
 void
 rb_gc_call_finalizer_at_exit()
 {
@@ -1772,14 +1824,9 @@ rb_gc_call_finalizer_at_exit()
 
     /* run finalizers */
     if (need_call_final) {
-	if (deferred_final_list) {
-	    p = deferred_final_list;
-	    while (p) {
-		RVALUE *tmp = p;
-		p = p->as.free.next;
-		run_final((VALUE)tmp);
-	    }
-	}
+	p = deferred_final_list;
+	deferred_final_list = 0;
+	finalize_list(p);
 	for (i = 0; i < heaps_used; i++) {
 	    p = heaps[i].slot; pend = p + heaps[i].limit;
 	    while (p < pend) {
@@ -1818,14 +1865,14 @@ rb_gc_call_finalizer_at_exit()
 /*
  *  call-seq:
  *     ObjectSpace._id2ref(object_id) -> an_object
- *  
+ *
  *  Converts an object id to a reference to the object. May not be
  *  called on an object id passed as a parameter to a finalizer.
- *     
+ *
  *     s = "I am a string"                    #=> "I am a string"
  *     r = ObjectSpace._id2ref(s.object_id)   #=> "I am a string"
  *     r == s                                 #=> true
- *     
+ *
  */
 
 static VALUE
