@@ -2,7 +2,7 @@
 
   re.c -
 
-  $Author: nobu $
+  $Author: matz $
   created at: Mon Aug  9 18:24:49 JST 1993
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -105,7 +105,7 @@ rb_memsearch(x0, m, y0, n)
     int d;
     unsigned long hx, hy;
 
-#define KR_REHASH(a, b, h) (((h) << 1) - ((a)<<d) + (b))
+#define KR_REHASH(a, b, h) (((h) << 1) - (((unsigned long)(a))<<d) + (b))
 
     if (m > n) return -1;
     s = y; e = s + n - m;
@@ -641,7 +641,9 @@ make_regexp(s, len, flags)
     err = re_compile_pattern(s, len, rp);
 
     if (err != NULL) {
+	re_free_pattern(rp);
 	rb_reg_raise(s, len, err, 0);
+	return 0;
     }
     return rp;
 }
@@ -678,6 +680,7 @@ match_alloc(klass)
     return (VALUE)match;
 }
 
+/* :nodoc: */
 static VALUE
 match_init_copy(obj, orig)
     VALUE obj, orig;
@@ -1523,8 +1526,8 @@ rb_reg_match(re, str)
  *     
  *     a = "HELLO"
  *     case a
- *     when /^a-z*$/; print "Lower case\n"
- *     when /^A-Z*$/; print "Upper case\n"
+ *     when /^[a-z]*$/; print "Lower case\n"
+ *     when /^[A-Z]*$/; print "Upper case\n"
  *     else;            print "Mixed case\n"
  *     end
  *     
@@ -1627,7 +1630,7 @@ rb_reg_match_m(re, str)
  *  options are propagated, and new options may not be specified (a change as of
  *  Ruby 1.8). If <i>options</i> is a <code>Fixnum</code>, it should be one or
  *  more of the constants <code>Regexp::EXTENDED</code>,
- *  <code>Regexp::IGNORECASE</code>, and <code>Regexp::POSIXLINE</code>,
+ *  <code>Regexp::IGNORECASE</code>, and <code>Regexp::MULTILINE</code>,
  *  <em>or</em>-ed together. Otherwise, if <i>options</i> is not
  *  <code>nil</code>, the regexp will be case insensitive. The <i>lang</i>
  *  parameter enables multibyte support for the regexp: `n', `N' = none, `e',
@@ -1681,8 +1684,6 @@ rb_reg_initialize_m(argc, argv, self)
 	len = RREGEXP(argv[0])->len;
     }
     else {
-	s = StringValuePtr(argv[0]);
-	len = RSTRING(argv[0])->len;
 	if (argc >= 2) {
 	    if (FIXNUM_P(argv[1])) flags = FIX2INT(argv[1]);
 	    else if (RTEST(argv[1])) flags = RE_OPTION_IGNORECASE;
@@ -1708,6 +1709,8 @@ rb_reg_initialize_m(argc, argv, self)
 		break;
 	    }
 	}
+	s = StringValuePtr(argv[0]);
+	len = RSTRING(argv[0])->len;
     }
     rb_reg_initialize(self, s, len, flags);
     return self;
@@ -1801,11 +1804,12 @@ rb_reg_quote(str)
 
 /*
  *  call-seq:
- *     Regexp.escape(str)   => new_str
- *     Regexp.quote(str)    => new_str
+ *     Regexp.escape(str)   => a_str
+ *     Regexp.quote(str)    => a_str
  *  
  *  Escapes any characters that would have special meaning in a regular
- *  expression. For any string,
+ *  expression. Returns a new escaped string, or self if no characters are
+ *  escaped.  For any string,
  *  <code>Regexp.escape(<i>str</i>)=~<i>str</i></code> will be true.
  *     
  *     Regexp.escape('\\*?{}.')   #=> \\\\\*\?\{\}\.
@@ -1971,6 +1975,7 @@ rb_reg_s_union(argc, argv)
     }
 }
 
+/* :nodoc: */
 static VALUE
 rb_reg_init_copy(copy, re)
     VALUE copy, re;
