@@ -2,8 +2,8 @@
 
   variable.c -
 
-  $Author: ocean $
-  $Date: 2004/07/15 11:42:51 $
+  $Author: nobu $
+  $Date: 2005/09/28 15:58:32 $
   created at: Tue Apr 19 23:55:15 JST 1994
 
   Copyright (C) 1993-2003 Yukihiro Matsumoto
@@ -98,7 +98,7 @@ fc_i(key, value, res)
 	    arg.klass = res->klass;
 	    arg.track = value;
 	    arg.prev = res;
-	    st_foreach(RCLASS(value)->iv_tbl, fc_i, (st_data_t)&arg);
+	    st_foreach_safe(RCLASS(value)->iv_tbl, fc_i, (st_data_t)&arg);
 	    if (arg.path) {
 		res->path = arg.path;
 		return ST_STOP;
@@ -124,7 +124,7 @@ find_class_path(klass)
     arg.track = rb_cObject;
     arg.prev = 0;
     if (RCLASS(rb_cObject)->iv_tbl) {
-	st_foreach(RCLASS(rb_cObject)->iv_tbl, fc_i, (st_data_t)&arg);
+	st_foreach_safe(RCLASS(rb_cObject)->iv_tbl, fc_i, (st_data_t)&arg);
     }
     if (arg.path == 0) {
 	st_foreach(rb_class_tbl, fc_i, (st_data_t)&arg);
@@ -196,6 +196,7 @@ rb_class_path(klass)
     }
     else {
 	char *s = "Class";
+	size_t len;
 
 	if (TYPE(klass) == T_MODULE) {
 	    if (rb_obj_class(klass) == rb_cModule) {
@@ -205,8 +206,9 @@ rb_class_path(klass)
 		s = rb_class2name(RBASIC(klass)->klass);
 	    }
 	}
-	path = rb_str_new(0, 2 + strlen(s) + 3 + 2 * SIZEOF_LONG + 1);
-	sprintf(RSTRING(path)->ptr, "#<%s:0x%lx>", s, klass);
+	len = 2 + strlen(s) + 3 + 2 * SIZEOF_LONG + 1;
+	path = rb_str_new(0, len);
+	snprintf(RSTRING(path)->ptr, len+1, "#<%s:0x%lx>", s, klass);
 	RSTRING(path)->len = strlen(RSTRING(path)->ptr);
 	rb_ivar_set(klass, tmp_classpath, path);
 
@@ -842,9 +844,10 @@ rb_generic_ivar_table(obj)
 }
 
 static VALUE
-generic_ivar_get(obj, id)
+generic_ivar_get(obj, id, warn)
     VALUE obj;
     ID id;
+    int warn;
 {
     st_table *tbl;
     VALUE val;
@@ -856,8 +859,9 @@ generic_ivar_get(obj, id)
 	}
       }
     }
-
-    rb_warning("instance variable %s not initialized", rb_id2name(id));
+    if (warn) {
+	rb_warning("instance variable %s not initialized", rb_id2name(id));
+    }
     return Qnil;
 }
 
@@ -958,7 +962,7 @@ rb_mark_generic_ivar_tbl()
 {
     if (!generic_iv_tbl) return;
     if (special_generic_ivar == 0) return;
-    st_foreach(generic_iv_tbl, givar_i, 0);
+    st_foreach_safe(generic_iv_tbl, givar_i, 0);
 }
 
 void
@@ -1008,13 +1012,12 @@ ivar_get(obj, id, warn)
 	break;
       default:
 	if (FL_TEST(obj, FL_EXIVAR) || rb_special_const_p(obj))
-	    return generic_ivar_get(obj, id);
+	    return generic_ivar_get(obj, id, warn);
 	break;
     }
-    if (warn && ruby_verbose) {
+    if (warn) {
 	rb_warning("instance variable %s not initialized", rb_id2name(id));
     }
-
     return Qnil;
 }
 
@@ -1118,7 +1121,7 @@ rb_obj_instance_variables(obj)
       case T_CLASS:
       case T_MODULE:
 	if (ROBJECT(obj)->iv_tbl) {
-	    st_foreach(ROBJECT(obj)->iv_tbl, ivar_i, ary);
+	    st_foreach_safe(ROBJECT(obj)->iv_tbl, ivar_i, ary);
 	}
 	break;
       default:
@@ -1127,7 +1130,7 @@ rb_obj_instance_variables(obj)
 	    st_table *tbl;
 
 	    if (st_lookup(generic_iv_tbl, obj, (st_data_t *)&tbl)) {
-		st_foreach(tbl, ivar_i, ary);
+		st_foreach_safe(tbl, ivar_i, ary);
 	    }
 	}
 	break;
@@ -1289,7 +1292,7 @@ rb_autoload(mod, id, file)
 	tbl = check_autoload_table(av);
     }
     else {
-	av = Data_Wrap_Struct(rb_cData, rb_mark_tbl, st_free_table, 0);
+	av = Data_Wrap_Struct(0 , rb_mark_tbl, st_free_table, 0);
 	st_add_direct(tbl, autoload, av);
 	DATA_PTR(av) = tbl = st_init_numtable();
     }
@@ -1326,7 +1329,7 @@ autoload_delete(mod, id)
     return (NODE *)load;
 }
 
-void
+VALUE
 rb_autoload_load(klass, id)
     VALUE klass;
     ID id;
@@ -1335,9 +1338,9 @@ rb_autoload_load(klass, id)
     NODE *load = autoload_delete(klass, id);
 
     if (!load || !(file = load->nd_lit) || rb_provided(RSTRING(file)->ptr)) {
-	const_missing(klass, id);
+	return Qfalse;
     }
-    rb_require_safe(file, load->nd_nth);
+    return rb_require_safe(file, load->nd_nth);
 }
 
 static VALUE
@@ -1355,7 +1358,7 @@ autoload_file(mod, id)
     }
     file = ((NODE *)load)->nd_lit;
     Check_Type(file, T_STRING);
-    if (!RSTRING(file)->ptr) {
+    if (!RSTRING(file)->ptr || !*RSTRING(file)->ptr) {
 	rb_raise(rb_eArgError, "empty file name");
     }
     if (!rb_provided(RSTRING(file)->ptr)) {
@@ -1403,7 +1406,7 @@ rb_const_get_0(klass, id, exclude, recurse)
     while (tmp) {
 	while (RCLASS(tmp)->iv_tbl && st_lookup(RCLASS(tmp)->iv_tbl,id,&value)) {
 	    if (value == Qundef) {
-		rb_autoload_load(tmp, id);
+		if (!RTEST(rb_autoload_load(tmp, id))) break;
 		continue;
 	    }
 	    if (exclude && tmp == rb_cObject && klass != rb_cObject) {
@@ -1511,7 +1514,7 @@ rb_mod_const_at(mod, data)
 	tbl = st_init_numtable();
     }
     if (RCLASS(mod)->iv_tbl) {
-	st_foreach(RCLASS(mod)->iv_tbl, sv_i, (st_data_t)tbl);
+	st_foreach_safe(RCLASS(mod)->iv_tbl, sv_i, (st_data_t)tbl);
     }
     return tbl;
 }
@@ -1875,7 +1878,7 @@ rb_mod_class_variables(obj)
 
     for (;;) {
 	if (RCLASS(obj)->iv_tbl) {
-	    st_foreach(RCLASS(obj)->iv_tbl, cv_i, ary);
+	    st_foreach_safe(RCLASS(obj)->iv_tbl, cv_i, ary);
 	}
 	obj = RCLASS(obj)->super;
 	if (!obj) break;
